

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

public class Formulae {

	public static void printReservations(Fraction[] reservations) {

		// System.out.println();
		int i = 0;
		while (i < reservations.length) {
			System.out.printf(" %.2f ", reservations[i].asDouble());
			i++;
		}
		System.out.println();
	}

	public static Fraction sumOfCapacity(Fraction[] capacity) {

		Fraction sum = new Fraction(0, 1);

		int i = 0;
		while (i < capacity.length) {
			sum = sum.plus(capacity[i]);
			i++;

		}
		return sum;
	}

	public static Fraction[] optimumReservations(Fraction alpha[]) {

		Fraction roOptimum[] = new Fraction[alpha.length];

		int i = 0;
		while (i < alpha.length) {
			roOptimum[i] = alpha[i].dividedBy(sumOfCapacity(alpha));
			i++;
		}
		return roOptimum;

	}
	
	public static Fraction[] evenAttackersTraffic(Fraction ro[], Fraction sumB) {

		Fraction evenBeta[] = new Fraction[ro.length];

		int i = 0;
		while (i < ro.length) {
			evenBeta[i] = sumB.times(ro[i]);
			i++;
		}
		return evenBeta;

	}

	public static Fraction[] optimumAttackersStrategy1(Fraction[] alpha,
			Fraction[] beta) {

		Fraction betaOptimum[] = new Fraction[alpha.length];
		int i = 0;

		while (i < alpha.length) {
			betaOptimum[i] = sumOfCapacity(beta).times(
					alpha[i].dividedBy(sumOfCapacity(alpha)));
			i++;
		}
		return betaOptimum;
	}

	public static Fraction[] optimumAttackersStrategy2(Fraction[] alpha,
			Fraction[] beta, Fraction[] ro) {

		Fraction[] betaOptimum2 = new Fraction[beta.length];
		Fraction subres = new Fraction(0, 1);
		Double help[] = new Double[beta.length];
		long l;

		int i = 0;
		while (i < beta.length) {
			help[i] = alpha[i].times(ro[i]).asDouble();
			help[i] = Math.sqrt(help[i]);
			l = new Double((help[i] * new Double(1000))).longValue();
			subres = subres.plus(new Fraction(l, 1000));
			i++;
		}
		// attackers adaptive best strategy
		i = 0;
		while (i < beta.length) {
			betaOptimum2[i] = new Fraction(new Double(
					(help[i] * new Double(100))).longValue(), 100);
			betaOptimum2[i] = betaOptimum2[i].dividedBy(subres).times(
					sumOfCapacity(alpha).plus(sumOfCapacity(beta)));
			betaOptimum2[i] = betaOptimum2[i].minus(alpha[i]);
			i++;
		}
		return betaOptimum2;
	}

	public static Fraction[] optimumAdaptiveReservations(Fraction[] alpha,
			Fraction[] beta) {

		Fraction[] c = new Fraction[alpha.length];
		Fraction roAdaptedOptimum[] = new Fraction[alpha.length];
		SortedMap<Fraction, Integer> sortedCoefs = new TreeMap<Fraction, Integer>();

		// calculate and sort values of coefficients
		for (int i = 0; i < alpha.length; i++) {
			c[i] = alpha[i];
			c[i] = c[i].dividedBy((alpha[i].plus(beta[i])));
			sortedCoefs.put(c[i], i);
		}
		// calculate roConcreteOptimum
		Set s = sortedCoefs.entrySet();
		// Using iterator in SortedMap
		Iterator iterator = s.iterator();
		while (iterator.hasNext()) {
			Map.Entry m = (Map.Entry) iterator.next();
			Fraction coeficient = (Fraction) m.getKey();
			Integer index = (Integer) m.getValue();
		}

		Fraction roConcreteSum = new Fraction(0, 1);
		while (!sortedCoefs.isEmpty()) {

			Fraction lastKey = sortedCoefs.lastKey();
			int indexOrder = sortedCoefs.get(lastKey);

			if (new Fraction(1, 1).minus(roConcreteSum).compareTo(
					alpha[indexOrder].plus(beta[indexOrder])) >= 0) {
				roAdaptedOptimum[indexOrder] = alpha[indexOrder]
						.plus(beta[indexOrder]);
			} else {
				roAdaptedOptimum[indexOrder] = new Fraction(1, 1)
						.minus(roConcreteSum);
			}
			roConcreteSum = roConcreteSum.plus(roAdaptedOptimum[indexOrder]);
			sortedCoefs.remove(lastKey);
		}
		return roAdaptedOptimum;
	}

	public static void defaultFormula(Fraction[] alpha, Fraction[] beta,
			Fraction[] ro) {

		Fraction[] cluster = new Fraction[alpha.length];
		Fraction total = new Fraction(0, 1);
		int i = 0;

		while (i < alpha.length) {
			cluster[i] = alpha[i].times(ro[i]);
			cluster[i] = cluster[i].dividedBy((alpha[i].plus(beta[i])));
			total = total.plus(cluster[i]);
			// System.out.printf(" %.2f", total.asDouble());
			i++;
		}
		// System.out.println();
		printReservations(cluster);
		System.out.printf("TOTAL: %.2f", total.asDouble());
		System.out.println();
	}

	public static void FormulaNoClusters(Fraction[] alpha, Fraction[] beta) {

		System.out.println();
		System.out.println("Result of NO CLUSTERS:");
		System.out.printf("TOTAL: %.2f", sumOfCapacity(alpha).dividedBy(
				sumOfCapacity(alpha).plus(sumOfCapacity(beta))).asDouble());
		System.out.println();

	}

}
