'''
Created on Mar 17, 2011

@author: sanja7s
'''
from Logger import logger
import Algorithms
import EvaluationFunctions
import math

def isNodeCluster(node, clusterSet):
    for cluster in clusterSet:
        if node['idNum']+node['rootDist'] == cluster['idNum']+cluster['rootDist']:
            return True
    return False

def getLeafFrequenciesOfNode(node, attribute='idNum'):
    
    leafFrequencies = {}
    
    if node['left'] != None:
        leftSubtreeFrequencies = getLeafFrequenciesOfNode(node['left'])
        for k in leftSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + leftSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = leftSubtreeFrequencies[k]
    if node['right'] != None:
        rightSubtreeFrequencies = getLeafFrequenciesOfNode(node['right'])
        for k in rightSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + rightSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = rightSubtreeFrequencies[k]
               
    if node['left'] == None and node['right'] == None:
        leafTotalCount = node['countGood'] + node['countNeutral'] + node['countEvil']
        if node[attribute]+node['rootDist'] in leafFrequencies:
            leafFrequencies[node[attribute]+node['rootDist']] += leafTotalCount
        else:
            leafFrequencies[node[attribute]+node['rootDist']] = leafTotalCount
    
    return leafFrequencies

def getLeafFrequenciesOfCluster(clusterNode, node, clusterSet, attribute='idNum'):
    
    leafFrequencies = {}
    
    """if clusterSet == None:
        logger.error("No cluster set, cannot calculate frequencies.")
        return
    if clusterNode not in clusterSet:
        logger.error("The node you are calculating the entropy for is not a cluster, cannot calculate frequencies..")
        return"""
        
    if node['left'] != None and isNodeCluster(node['left'],clusterSet) == False:
        leftSubtreeFrequencies = getLeafFrequenciesOfCluster(clusterNode, node['left'], clusterSet)
        for k in leftSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + leftSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = leftSubtreeFrequencies[k]
    if node['right'] != None and isNodeCluster(node['right'],clusterSet) == False:
        rightSubtreeFrequencies = getLeafFrequenciesOfCluster(clusterNode, node['right'], clusterSet)
        for k in rightSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + rightSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = rightSubtreeFrequencies[k]
    if node['left'] == None and node['right'] == None:
        if isNodeCluster(node,clusterSet) == False or clusterNode['idNum']+clusterNode['rootDist'] == node['idNum']+node['rootDist']:
            leafTotalCount = node['countGood'] + node['countNeutral'] + node['countEvil']
            if node[attribute] in leafFrequencies:
                leafFrequencies[node[attribute]] += leafTotalCount
            else:
                leafFrequencies[node[attribute]] = leafTotalCount
         
    return leafFrequencies
 
def getClusteringAvgSumEntropy(clusterSet, tree):
    
    entropy = 0
    leafFrequencies = {}
    
    if clusterSet == None:
        logger.error("No cluster set")
        return
    
    totalCount = tree.root['countGood'] + tree.root['countNeutral'] + tree.root['countEvil']
    #print('Total items in the dataset  ',totalCount)
    
    k = 0
    for cluster in clusterSet:
        k += 1
        #if cluster['idNum'] == 0:
            #print('yes')
        node = tree.getNodeFromId(cluster['idNum']+cluster['rootDist'],tree.root)
        #print('cluster',cluster['idNum'],cluster['rootDist'])
        if node == None:
            logger.error('The given cluster is not a node from the given tree. Cannot proceed.')
            return
        totalClusterCount = 0
        leafFrequencies = getLeafFrequenciesOfCluster(cluster, node, clusterSet)
        #print(leafFrequencies)
        for itemCount in leafFrequencies.values():
            totalClusterCount += itemCount
        #print(totalCount)
        #print(totalClusterCount)
        #print(calculateEntropy(leafFrequencies))
 
        entropy += calculateEntropy(leafFrequencies)*totalClusterCount/totalCount
    #print(k)
    return entropy   

def calculateEntropy(dataItemFrequencies):
    """
    Calculates the entropy of the given data set and its items' frequencies.
    """
    if dataItemFrequencies == False:
        logger.error("Empty frequencies list, cannot calculate entropy.")
        return
    
    dataEntropy = 0.0
    totalCount = 0
    
    for itemCount in dataItemFrequencies.values():
        totalCount += itemCount
    #print(totalCount)

    #Calculate the entropy of the data for the target attribute
    for itemCount in dataItemFrequencies.values():
        if itemCount == 0:
            logger.warning('Zero value in frequencies list, we calculate entropy as log 0 = 0.')
        else:
            dataEntropy += (-itemCount/totalCount) * math.log(itemCount/totalCount, 2) 
        
    return dataEntropy


def printClusteringAvgEntropiesPerThreshold(clustering, tree, min=1, max=25, step=1):
    """
    
    """
    print('Testing entropies of different threshold clusterings')
    for i in range(min, max, step):
        #print('Entropy of the clustering with threshold  ',i,' pct')
        clustering = Algorithms.HHHinspiredPerPercentage(tree.root, tree.root, i)    
        clusteringEntropy = EvaluationFunctions.getClusteringAvgSumEntropy(clustering,tree)
        print(i, "  ", clusteringEntropy)