'''
Created on Mar 17, 2011

@author: sanja7s
'''
from Logger import logger
import Algorithms
import math
from copy import deepcopy
from IPGenerator import getIPStringFromIPNum

def isNodeCluster(node, clusterSet):
    """
    Returns boolean answer whether the node is a cluster in the given cluster set.
    """
    for cluster in clusterSet:
        if node['idNum']+node['rootDist'] == cluster['idNum']+cluster['rootDist']:
            return True
    return False

def getLeafFrequenciesOfNode(node, attribute='idNum'):
    """
    Returns the list of all leaf ids with their frequencies in the subtree which has as the root - 'node'.
    """
    
    leafFrequencies = {}
    
    if node['left'] != None:
        leftSubtreeFrequencies = getLeafFrequenciesOfNode(node['left'])
        for k in leftSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + leftSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = leftSubtreeFrequencies[k]
    if node['right'] != None:
        rightSubtreeFrequencies = getLeafFrequenciesOfNode(node['right'])
        for k in rightSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + rightSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = rightSubtreeFrequencies[k]
               
    if node['left'] == None and node['right'] == None:
        leafTotalCount = node['countGood'] + node['countNeutral'] + node['countEvil']
        if node[attribute]+node['rootDist'] in leafFrequencies:
            leafFrequencies[node[attribute]+node['rootDist']] += leafTotalCount
        else:
            leafFrequencies[node[attribute]+node['rootDist']] = leafTotalCount
    
    return leafFrequencies

def getLeafFrequenciesOfCluster(clusterNode, node, clusterSet, attribute='idNum'):
    """
    Returns the list of leaf ids and their frequencies in the subtree which has as the root - 'node', but 'node' is a cluster. 
    """
    # Different from previous, since when node is a cluster - we need to omit its ancestors that belong to other clusters.
    leafFrequencies = {}
    
    if node['left'] != None and isNodeCluster(node['left'],clusterSet) == False:
        leftSubtreeFrequencies = getLeafFrequenciesOfCluster(clusterNode, node['left'], clusterSet)
        for k in leftSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + leftSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = leftSubtreeFrequencies[k]
    if node['right'] != None and isNodeCluster(node['right'],clusterSet) == False:
        rightSubtreeFrequencies = getLeafFrequenciesOfCluster(clusterNode, node['right'], clusterSet)
        for k in rightSubtreeFrequencies:
            if k in leafFrequencies:
                leafFrequencies[k] = leafFrequencies[k] + rightSubtreeFrequencies[k]
            else:
                leafFrequencies[k] = rightSubtreeFrequencies[k]
    if node['left'] == None and node['right'] == None:
        if isNodeCluster(node,clusterSet) == False or clusterNode['idNum']+clusterNode['rootDist'] == node['idNum']+node['rootDist']:
            leafTotalCount = node['countGood'] + node['countNeutral'] + node['countEvil']
            if node[attribute] in leafFrequencies:
                leafFrequencies[node[attribute]] += leafTotalCount
            else:
                leafFrequencies[node[attribute]] = leafTotalCount
         
    return leafFrequencies

def getClusteringEntropy(clusterSet, tree):
    '''
    Invokes a simple entropy calculation of entropy with clusters as the items of the set.
    '''
    clusterFrequencies = {}
    # we put in frequency list cluster counts, and for uniqueness we use as the key in the dictionary: cluster['idNum']+cluster['rootDist']. 
    for cluster in clusterSet:
        node = tree.getNodeFromId(cluster['idNum']+cluster['rootDist'],tree.root)
        if node == None:
            logger.error('The given cluster is not a node from the given tree. Cannot proceed.')
            return
        totalClusterCount = 0
        leafFrequencies = getLeafFrequenciesOfCluster(cluster, node, clusterSet)
        for itemCount in leafFrequencies.values():
            totalClusterCount += itemCount
        clusterFrequencies[cluster['idNum']+cluster['rootDist']] = totalClusterCount

    return(calculateEntropy(clusterFrequencies))
 
def getClusteringAvgSumEntropy(clusterSet, tree):
    """
    Calls getLeafFrequenciesOfCluster for each cluster, calculates inter-cluster entropies and then averages the sum entropy.
    """
    
    entropy = 0
    leafFrequencies = {}
    
    if clusterSet == None:
        logger.error("No cluster set")
        return
    
    totalCount = tree.root['countGood'] + tree.root['countNeutral'] + tree.root['countEvil']

    
    k = 0
    for cluster in clusterSet:
        k += 1
        node = tree.getNodeFromId(cluster['idNum']+cluster['rootDist'],tree.root)
        if node == None:
            logger.error('The given cluster is not a node from the given tree. Cannot proceed.')
            return
        totalClusterCount = 0
        leafFrequencies = getLeafFrequenciesOfCluster(cluster, node, clusterSet)

        for itemCount in leafFrequencies.values():
            totalClusterCount += itemCount

        entropy += calculateEntropy(leafFrequencies)*totalClusterCount/totalCount

    return entropy

def getRelativeClusteringMetricValue(clusterSet):
    
    metricValue = 0
    
    if clusterSet == None:
        logger.error("No cluster set")
        return
    
    k = 0
    for cluster in clusterSet:
        k += 1
        metricValue = metricValue + cluster['rootDist']
        
    return metricValue/k

def getClusteringRelativeMetricEntropy(clusterSet, tree):
     
    if clusterSet == None:
        logger.error("No cluster set")
        return
    
    entropyValue = 0
    clusterFrequencies = {}
    leafsFreq = getLeafFrequenciesOfNode(tree.root)
    totalEntropy = calculateEntropy(leafsFreq)
    totalCount = tree.root['countGood'] + tree.root['countNeutral'] + tree.root['countEvil']
    
    k = 0
    for cluster in clusterSet:
        clusterCount = 0
        k += 1
        node = tree.getNodeFromId(cluster['idNum']+cluster['rootDist'],tree.root)
        leafFrequencies = getLeafFrequenciesOfCluster(cluster, node, clusterSet)
        clusterEntropy = calculateEntropy(leafFrequencies)
        
        for itemCount in leafFrequencies.values():
            clusterCount += itemCount
        
        #variant 1
        # entropyValue += clusterEntropy/(32 - cluster['rootDist']+1)
        # this only makes meaning for me ... we want higher level clusters with smaller entropy inside of them
        # but do we want 
        entropyValue += cluster['rootDist']/(clusterEntropy + 1)*clusterCount
        
        #variant 1.1
        #if clusterEntropy != 0:
        #    entropyValue = entropyValue + cluster['rootDist']/clusterEntropy
        #else:
        #    entropyValue = entropyValue + cluster['rootDist']
            
        #variant 2
        #entropyValue += cluster['rootDist']*clusterEntropy*totalClusterCount/totalCount    
            
        #variant 2.1
        
        """ print(cluster['rootDist'], "distance " , clusterEntropy , " id " ,cluster['idNum'])  

        if clusterEntropy != 0:
            entropyValue += cluster['rootDist']/clusterEntropy*totalClusterCount/totalCount
        else:
            entropyValue += cluster['rootDist']*totalClusterCount/totalCount"""
            
        #variant 3
        #entropyValue +=  clusterCount * clusterEntropy * (32 -cluster['rootDist'])
        
        # variant 3.1
        """if clusterCount != 0:
            s = 2 ^ (32 -cluster['rootDist']) 
            entropyValue +=  clusterEntropy * s / clusterCount"""
            
        #variant Andrey1
        """if clusterCount != 0:
            s = 2 ^ (32 - cluster['rootDist'])
            clusterFrequencies[cluster['idNum'] + cluster['rootDist']] = s / clusterCount """
        
        #variant Andrey2
        #clusterFrequencies[cluster['idNum'] + cluster['rootDist']] = 32 - cluster['rootDist']
        
        
    return entropyValue/totalCount
    #return entropyValue
    #return calculateEntropy(clusterFrequencies)


def printRelativeMetricEntropyPerThreshold(tree, output, min=1, max=25, step=1, filename='relativeMetric.txt' ):
        
    k = 1
    
    if step < 1:
        k = step
        step = 1
        max = math.ceil(max / k)
        min = math.ceil(min / k)
    
    if output == 'screen':
        print('Testing relative metric of the clustering with threshold')
        for i in range(min, max, step):
            j = i * k
            clustering = Algorithms.HHHinspiredPerPercentage(tree.root, tree.root, j)
            
            metricValue = getClusteringRelativeMetricEntropy(clustering, tree)
            print('threshold ', j,' ',metricValue)
    elif output == 'file':
        if filename == None:
                logger.error("Print printRelativeClusteringMetricValuePerThreshold in a file, output without filename specified.")
                return False
        outfile = open(filename, 'w')
        outfile.write("Testing relative metric of the clustering with different thresholds:"+"\n")
        for i in range(min, max, step):
            j = i * k
            clustering = Algorithms.HHHinspiredPerPercentage(tree.root, tree.root, j)    
            metricValue = getClusteringRelativeMetricEntropy(clustering, tree)
            outfile.write(str(j) +"  "+ str(metricValue)+"\n")
        outfile.write("\n")
    else:
        logger.info("No output type \"" + output + "\" in printClusteringAvgEntropiesPerThreshold.")
        return False
        
    

def calculateEntropy(dataItemFrequencies):
    """
    Calculates the entropy of the given data set and its items' frequencies.
    """
    if dataItemFrequencies == False:
        logger.error("Empty frequencies list, cannot calculate entropy.")
        return
    
    dataEntropy = 0.0
    totalCount = 0
    
    for itemCount in dataItemFrequencies.values():
        totalCount += itemCount
 
    #Calculate the entropy of the data for the target attribute
    for itemCount in dataItemFrequencies.values():
        if itemCount == 0:
            logger.warning('Zero value in frequencies list, we calculate entropy as log 0 = 0.')
        else:
            dataEntropy += (-itemCount/totalCount) * math.log(itemCount/totalCount, 2) 
        
    return dataEntropy


def printClusteringAvgEntropiesPerThreshold(tree, output, min=1, max=25, step=1, filename='entropy.txt'):
    """
    This function is used for tests. It can print on screen or in the specified file - results of clustering entropies for given thresholds (between given limits
    and with given step of increase),
    but it is very slow since it needs to invoke HHHinspired algorithm for each different threshold.
    """
    k = 1
    
    if step < 1:
        k = step
        step = 1
        max = math.ceil(max / k)
        min = math.ceil(min / k)
        
    if output == 'screen':
        print('Testing entropies of different threshold clusterings')
        for i in range(min, max, step):
        #print('Entropy of the clustering with threshold  ',i,' pct')
            j = i * k
            clustering = Algorithms.HHHinspiredPerPercentage(tree.root, tree.root, j)    
            clusteringEntropy = getClusteringAvgSumEntropy(clustering,tree)
            print(j, "  ", clusteringEntropy)
    elif output == 'file':
        if filename == None:
                logger.error("Print ClusteringAvgEntropiesPerThreshold in a file, output without filename specified.")
                return False
        outfile = open(filename, 'w')
        outfile.write("Results of testing entropies of different threshold clusterings:"+"\n")
        for i in range(min, max, step):
            j = i * k
            clustering = Algorithms.HHHinspiredPerPercentage(tree.root, tree.root, j)    
            clusteringEntropy = getClusteringAvgSumEntropy(clustering,tree)
            outfile.write(str(j) +"  "+ str(clusteringEntropy)+"\n")
        outfile.write("\n")
    else:
        logger.info("No output type \"" + output + "\" in printClusteringAvgEntropiesPerThreshold.")
        return False
    
def createHTBfilter(tree, sortedclusternodes, filename, capacity = 500):
    
    leafFreq = getLeafFrequenciesOfNode(tree.root)
    total = 0
    for itemCount in leafFreq.values():
            total += itemCount
            
    sortedclusternodes = deepcopy(sortedclusternodes)
    
    HTBfile = open(filename, 'w')
    HTBfile.write("# HTB filter for the clustering"+"\n")
    HTBfile.write("\n")
    HTBfile.write("tc qdisc add dev eth4 handle 1: root htb default 10"+"\n")
    HTBfile.write("\n"+"\n")
    HTBfile.write("tc class add dev eth4 parent 1: classid 1:1 htb rate "+ str(capacity) + "kbps ceil "+ str(capacity) + "kbps"+"\n")
    HTBfile.write("\n"+"\n")
    
    print(total)
    
    clusterid = {}
    
    total1 = 0
    for node in sortedclusternodes:
        if (node['idNum']+node['rootDist']) != 0:
            total1 += node['countGood'] + node['countNeutral'] + node['countEvil']
            
    classid = 10
    for node in sortedclusternodes:
        if node['idNum']+node['rootDist'] == 0:
            count = total - total1
            rate = math.floor(capacity * count / total)
            clusterid[node['idNum']+node['rootDist']] = 10
            HTBfile.write("tc class add dev eth4 parent 1:1 classid 1:10 htb rate "+str(rate)+"kbps ceil "+ str(capacity) + "kbps"+"\n")
        else:
            classid += 1
            rate = math.floor(capacity * (node['countGood'] + node['countNeutral'] + node['countEvil']) / total)
            clusterid[node['idNum']+node['rootDist']] = classid
            HTBfile.write("tc class add dev eth4 parent 1:1 classid 1:"+ str(classid)+" htb rate "+str(rate)+"kbps ceil "+ str(capacity) + "kbps"+"\n")
    
    HTBfile.write("\n")
    HTBfile.write("\n")
    reversesortedclusternodes = sorted(sortedclusternodes, key=lambda node: node['rootDist'], reverse=True)
    for node in reversesortedclusternodes:
        classid = clusterid[node['idNum']+node['rootDist']]
        if classid != 10:
            srcIP = getIPStringFromIPNum(node['idNum']) + "/" + str(node['rootDist'])
            HTBfile.write("tc filter add dev eth4 parent 1: prio 1 protocol ip u32 match ip src "+srcIP+" flowid 1:"+str(classid)+"\n")
        
    HTBfile.write("\n")
    HTBfile.close()
        
        
def evaluateClusteringLevel(clustering):
       
    total = 0
    cluster_count = 0
    
    for cluster in clustering:
        total += cluster['rootDist'] * (cluster['countGood'] + cluster['countNeutral'] + cluster['countEvil'])
        cluster_count += 1
    
    #print (total)
    #print (cluster_count)
    return total / cluster_count
    